#include <SPI.h>
#include "ssd1306SPI.h"
#include "font8x16.h"
#include "bitmapdata.h"

// Declaration for SSD1306 display connected using software SPI (default case):
#define OLED_CS    10   // Slave Select(Chip select)
#define OLED_DC     6   // SSD1306 Data/Command
#define OLED_RESET  7   // SSD1306 Reset

// --- SSD1306 COMMADN : I2Cで使ったのと同じ ---
static uint8_t oledInitCommands[] = { // SSD1306の初期化コマンド群
  0xC8,         // 上下の描画方向
  0xA1,         // 左右の描画方向
  0xA8, 0x3F,   // 画面の解像度
  0xD3, 0x00,   // 縦方向のオフセット
  0x40,         // 縦方向の描画開始位置
  0xA6,         // 表示方法 - A7にすると画面反転
  0x8D, 0x14    // OLEDパネルの供給電圧と思われる 14hに設定しないとダメらしい
};

// 初期化
void SSD1306SPI::init() {
  // --- SOFTWARE SPI ピン設定
  pinMode(OLED_CS, OUTPUT);   // Sleva Select
  pinMode(OLED_DC, OUTPUT);   // SSD1306 Set data/command pin as output

  //SPI初期化
  digitalWrite(OLED_CS, LOW);
  SPI.begin();//SPIを初期化、SCK、MOSI、SSの各ピンの動作は出力、SCK、MOSIはLOW、SSはHIGH
  SPI.setDataMode(SPI_MODE0);//SPIモードを「0」に設定・CPOL(クロック位相)=0,CPHA(クロック極性)=0
  SPI.setBitOrder(MSBFIRST);//SPI送受信用のビットオーダーを「MSBFIRST」に設定

  // --- Reset SSD1306
  pinMode(OLED_RESET, OUTPUT);// SSD1306 RESET
  digitalWrite(OLED_RESET, HIGH);
  delay(1);                       // VDD goes high at start, pause for 1 ms
  digitalWrite(OLED_RESET, LOW);  // Bring reset low
  delay(10);                      // Wait 10 ms
  digitalWrite(OLED_RESET, HIGH); // Bring out of reset
  delay(100) ; 

  // --- SSD1306 初期化
  // OLED設定コマンド送信
  oledSendCommands(sizeof(oledInitCommands), oledInitCommands) ;  // 初期化
  oledSendCommand(0xA4) ; // 描画ON
  oledSendCommand(0xAF) ; // 画面表示
  delay(100) ; 

  curX = 0 ;
  curY = 0 ;
  fontColor = MONOOLED_WHITE ;

  cls() ;
  dispSwitch(true) ;
}

// 画面消去
void SSD1306SPI::cls() {
  // Franmeバッファ消去
  memset(DisplayBuff, 0, sizeof(DisplayBuff)) ;
  clsOLED() ;
}

// スクロール
void SSD1306SPI::scroll() {
  memcpy(DisplayBuff, &DisplayBuff[OLED_WIDTH], sizeof(DisplayBuff) - OLED_WIDTH ) ;
  memset(&DisplayBuff[sizeof(DisplayBuff) - OLED_WIDTH ], 0, OLED_WIDTH) ;
  oledDispFrame( ) ;
}

void SSD1306SPI::clsOLED() {
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト
  for (uint8_t m = 0; m < OLED_PAGES ; m++) {
    // 表示開始位置を指定
    digitalWrite(OLED_DC, LOW); // Command mode
    SPI.transfer(0xB0 | (m & 0x07));
    SPI.transfer(0x10);
    SPI.transfer(0x00);

    // 表示データ転送
    // １バイトづつ書き込みだと折角のペリフェラルが勿体ないので
    // バッファを介して１セグメント分づつ送信している
    digitalWrite(OLED_DC, HIGH); // Data mode
    memset(SPIBuf, 0x00, OLED_WIDTH) ;
    SPI.transfer(SPIBuf, OLED_WIDTH);
  }
  // 表示開始位置を指定
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(0xB0);
  SPI.transfer(0x10);
  SPI.transfer(0x00);

  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// 表示日表示切替
void SSD1306SPI::dispSwitch(boolean  sw) {
  if (sw) {
    oledSendCommand(0xAF) ; // 画面表示
  } else {
    oledSendCommand(0xAE) ; // 画面非表示
  }
}

void SSD1306SPI::Print(char *str){
  Print(curX,curY,str) ;
}

void SSD1306SPI::Println(char *str,bool flagScroll) {
  Print(str) ;
  curX = 0 ;
  curY ++ ;
  if (curY >= OLED_PAGES) {
    if (flagScroll) {
      scroll( ) ;
      curY = OLED_PAGES -1 ;
    } else {
      curY %= OLED_PAGES ;
    }
  }
}

void SSD1306SPI::locate(int x , int y) {
  curX = x ;
  curY = y ;
}

// 文字列表示
void SSD1306SPI::Print(int x, int y , char *str) {
  locate(x,y) ;
  int strLen = strlen(str) ;
  for (int i=0;i<strLen;i++) {
    DrowFont((uint8_t)str[i],curX,curY,fontColor,false) ; 
    curX ++ ;
  }
  oledDispFrame( ) ;
}

void SSD1306SPI::DrowFont(uint8_t code , uint8_t x, uint8_t y, uint8_t color,bool flagDisp) {
  int pos = (code - 0x20) * 16 ;
  for (int i=0;i<2;i++) {
    for (int j=0;j<8;j++) {
      int offextX = x * 8 + j ;
      uint8_t bitMap = pgm_read_byte(ssd1306xled_font8x16 + pos);
      switch (color) {
        case MONOOLED_WHITE:
          DisplayBuff[offextX + y * OLED_WIDTH] = bitMap;
          break;
        case MONOOLED_BLACK:
          DisplayBuff[offextX + y * OLED_WIDTH] = ~bitMap;
          break;
        case MONOOLED_INVERSE:
          DisplayBuff[offextX + y  * OLED_WIDTH] ^= bitMap;
          break;
      }
      pos ++ ;
    }
    y ++ ;
  }
  if (flagDisp) {
    oledDispFrame( ) ;
  }
}

// 7セグメント調フォント表示
void SSD1306SPI::disp7SEG(int offsetX , int offsetY , int num , uint8_t color , bool flagDisp) {
  if (num <= 18) {
    if (num < 0) { num = 10 ; }
    bool flag7[7] ;
    for (int i=0;i<7;i++) {
      flag7[i] = addMap[num][i] ; 
    }
    int pos = 0 ;
    for (int j = 0; j < 4 ; j++) {
      int y = offsetY + j ;
      for (int i = 0; i < 16 ; i++) {
        int x = offsetX * 8  + i ;
        uint8_t bitData = 0x00 ;
        if (flag7[0]) { bitData |= pgm_read_word(SEGa + pos) ; }  // a
        if (flag7[1]) { bitData |= pgm_read_word(SEGb + pos) ; }  // b
        if (flag7[2]) { bitData |= pgm_read_word(SEGc + pos) ; }  // c
        if (flag7[3]) { bitData |= pgm_read_word(SEGd + pos) ; }  // d
        if (flag7[4]) { bitData |= pgm_read_word(SEGe + pos) ; }  // e
        if (flag7[5]) { bitData |= pgm_read_word(SEGf + pos) ; }  // f
        if (flag7[6]) { bitData |= pgm_read_word(SEGg + pos) ; }  // g
        switch (color) {
          case MONOOLED_WHITE:
            DisplayBuff[x + y * OLED_WIDTH] = bitData;
            break;
          case MONOOLED_BLACK:
            DisplayBuff[x + y * OLED_WIDTH] = ~bitData;
            break;
          case MONOOLED_INVERSE:
            DisplayBuff[x + y * OLED_WIDTH] ^= bitData;
            break;
        }
        pos ++ ;
      }
    }
  } else {
    uint8_t *bitMap ;
    switch(num) {
      case 19: bitMap = SEGdot ; break ; // DOT
      case 20: bitMap = SEGtemp ; break ; // 温度計
      case 21: bitMap = SEGoc ; break ; // ℃
      case 22: bitMap = SEGhum ; break ; // 湿度計
      case 23: bitMap = SEGp ; break ; // ％
      case 24: bitMap = SEGcoron ; break ; // ：
    }
    int pos = 0 ;
    for (int j = 0; j < 4 ; j++) {
      int y = offsetY + j ;
      for (int i = 0; i < 16 ; i++) {
        int x = offsetX * 8  + i ;
        uint8_t bitData = pgm_read_byte(bitMap + pos) ; // pgm_read_byte / pgm_read_word
        switch (color) {
          case MONOOLED_WHITE:
            DisplayBuff[x + y * OLED_WIDTH] |= bitData;
            break;
          case MONOOLED_BLACK:
            DisplayBuff[x + y * OLED_WIDTH] &= ~bitData;
            break;
          case MONOOLED_INVERSE:
            DisplayBuff[x + y * OLED_WIDTH] ^= bitData;
            break;
        }
        pos ++ ;
      }
    } 
  }
  if (flagDisp) {
    oledDispFrame( ) ;
  }
}

// OLEDに1バイトコマンドを送信
void SSD1306SPI::oledSendCommand(uint8_t command) {
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// OLEDに2バイトコマンドを送信
void SSD1306SPI::oledSendCommand2(uint8_t command1, uint8_t command2) {
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(command1);
  SPI.transfer(command2);
  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// OLEDに3バイトコマンドを送信
void SSD1306SPI::oledSendCommand3(uint8_t command1, uint8_t command2, uint8_t command3) {
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(command1);
  SPI.transfer(command2);
  SPI.transfer(command3);
  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// OLEDに複数のコマンドを送信
void SSD1306SPI::oledSendCommands(int commandLen, uint8_t commands[]) {
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(commands, commandLen);
  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// 128x32ドットＯＬＥＤにビットマップを書き込み
void SSD1306SPI::oledBitMap(int col , int row , int width , uint8_t *bitMap) {
  int offsetX = col * 8 ;
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト

  // 表示開始位置を指定
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(0xB0 | (row & 0x07));
  SPI.transfer(0x10 | (offsetX & 0xf0) >> 4 );
  SPI.transfer(offsetX & 0x0F);

  // 表示データ転送
  // Arduinoだとメモリが足りなくて１画面分を一気に転送する事ができない
  // なので、１セグメント分づつ送信している
  digitalWrite(OLED_DC, HIGH); // Data mode
  SPI.transfer(bitMap, width);

  // 表示開始位置を指定
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(0xB0);
  SPI.transfer(0x10);
  SPI.transfer(0x00);

  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// BITパターン表示
void SSD1306SPI::oledBitMap(int offsetX , int offsetY , int width,int hight, uint8_t *bitMap) {
  int pos = 0 ;
  for (int m = 0; m < hight; m++) {
    memcpy(SPIBuf, bitMap , OLED_WIDTH) ;
    oledBitMap(offsetX , offsetY , width , SPIBuf) ;
    bitMap += width ;
    offsetY ++ ;
  }
}

// Fraebuff表示
void SSD1306SPI::oledDispFrame( ) {
  uint8_t *dispPos = DisplayBuff ;
  int offsetY = 0 ;
//  dispSwitch(false) ;
  for (int m = 0; m < OLED_PAGES; m++) {
    memcpy(SPIBuf,dispPos,OLED_WIDTH) ;
    oledBitMap(0 , offsetY , OLED_WIDTH , SPIBuf) ;
    dispPos += OLED_WIDTH ;
    offsetY ++ ;
  }
//  dispSwitch(true) ;
}

// ▼▼▼▼▼▼ Adafruit_GFX_Libraryからパクった描画関数
// ローテーションなし、Windowクリップなし、描画フレームへの書き込み限定にして、で不要部分を削ってます

void SSD1306SPI::drawPixel(int x, int y, uint8_t color,bool flagDisp) {
  if ((x >= 0) && (x < OLED_WIDTH) && (y >= 0) && (y < OLED_HEIGHT)) {
    switch (color) {
      case MONOOLED_WHITE:
        DisplayBuff[x + (y / 8) * OLED_WIDTH] |= (1 << (y & 7));
        break;
      case MONOOLED_BLACK:
        DisplayBuff[x + (y / 8) * OLED_WIDTH] &= ~(1 << (y & 7));
        break;
      case MONOOLED_INVERSE:
        DisplayBuff[x + (y / 8) * OLED_WIDTH] ^= (1 << (y & 7));
        break;
    }
  }
  if (flagDisp) {
    oledDispFrame( ) ;
  }
}

void SSD1306SPI::drawLine(int x0, int y0, int x1, int y1, uint8_t color, bool flagDisp) {
  int16_t steep = abs(y1 - y0) > abs(x1 - x0);
  if (steep) {
    _swap_int16_t(x0, y0);
    _swap_int16_t(x1, y1);
  }

  if (x0 > x1) {
    _swap_int16_t(x0, x1);
    _swap_int16_t(y0, y1);
  }

  int16_t dx, dy;
  dx = x1 - x0;
  dy = abs(y1 - y0);

  int16_t err = dx / 2;
  int16_t ystep;

  if (y0 < y1) {
    ystep = 1;
  } else {
    ystep = -1;
  }

  for (; x0 <= x1; x0++) {
    if (steep) {
      drawPixel(y0, x0, color , false);
    } else {
      drawPixel(x0, y0, color , false);
    }
    err -= dy;
    if (err < 0) {
      y0 += ystep;
      err += dx;
    }
  }

  if (flagDisp) {
    oledDispFrame( ) ;
  }
}

void SSD1306SPI::fillRect(int x, int y, int w, int h, uint8_t color, bool flagDisp) {
  for (int16_t i = x; i < x + w; i++) {
    drawLine(i, y, i , y + h -1 , color , false);
  }

  if (flagDisp) {
    oledDispFrame( ) ;
  }
}

void SSD1306SPI::drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint8_t color, bool flagDisp) {
  int xe = x + w - 1 ;
  int ye = y + h - 1 ;
  drawLine(x, y, xe,y ,color,false);
  drawLine(x, y, x ,ye,color,false);
  drawLine(xe,y, xe,ye,color,false);
  drawLine(x ,ye,xe,ye,color,false);

  if (flagDisp) {
    oledDispFrame( ) ;
  }
}

void SSD1306SPI::drawCircle(int16_t x0, int16_t y0, int16_t r, uint8_t color, bool flagDisp) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;

  drawPixel(x0, y0 + r, color,false);
  drawPixel(x0, y0 - r, color,false);
  drawPixel(x0 + r, y0, color,false);
  drawPixel(x0 - r, y0, color,false);

  while (x < y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;

    drawPixel(x0 + x, y0 + y, color,false);
    drawPixel(x0 - x, y0 + y, color,false);
    drawPixel(x0 + x, y0 - y, color,false);
    drawPixel(x0 - x, y0 - y, color,false);
    drawPixel(x0 + y, y0 + x, color,false);
    drawPixel(x0 - y, y0 + x, color,false);
    drawPixel(x0 + y, y0 - x, color,false);
    drawPixel(x0 - y, y0 - x, color,false);
  }

  if (flagDisp) {
    oledDispFrame( ) ;
  }
}

void SSD1306SPI::fillCircle(int16_t x0, int16_t y0, int16_t r,uint8_t color, bool flagDisp) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;
  int16_t px = x;
  int16_t py = y;
  int16_t delta = 1 ;
  
  drawLine(x0, y0 - r , x0 , y0 + r , color,false);

  while (x < y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;
    // These checks avoid double-drawing certain lines, important
    // for the SSD1306 library which has an INVERT drawing mode.
    if (x < (y + 1)) {
        drawLine(x0 + x , y0 - y , x0 + x , y0 + y + delta -1 , color,false);
        drawLine(x0 - x , y0 - y , x0 - x , y0 + y + delta -1 , color,false);
    }
    if (y != py) {
        drawLine(x0 + py , y0 - px , x0 + py , y0 + px + delta -1 , color,false);
        drawLine(x0 - py , y0 - px , x0 - py , y0 + px + delta -1 , color,false);
      py = y;
    }
    px = x;
  }

  if (flagDisp) {
    oledDispFrame( ) ;
  }
}
